#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" This is the execution manager. """

import logging
from typing import Dict, List, Optional

import transaction
from messaging.messenger import MessageSender
from messaging.router import Router, on_message

from workspaces.capability.enums import CapabilityRequestState, QueueState
from workspaces.capability.message_architect import CapabilityMessageArchitect
from workspaces.capability.schema_interfaces import CapabilityExecutionIF
from workspaces.capability.services.interfaces import (
    CapabilityInfoIF,
    ExecutionManagerIF,
)
from workspaces.capability.services.queue_manager import CapabilityQueueManager
from workspaces.system.schema import AbstractFile
from workspaces.workflow.services.interfaces import WorkflowServiceIF

# pylint: disable=E0401, W1203


logger = logging.getLogger(__name__)


class ExecutionManager(ExecutionManagerIF):
    """
    Class that manages the capability execution system. It's duties include:
    - Creating new executions
    - Managing existing execution state
    - Handling capability events and updating the state of the system accordingly
    """

    def __init__(
        self,
        capability_info: CapabilityInfoIF,
        workflow_service: WorkflowServiceIF,
        router: Router,
        messenger: MessageSender,
    ):
        self.capability_info = capability_info
        self.workflow_service = workflow_service
        self.messenger = messenger
        self.queue_manager = CapabilityQueueManager(self.capability_info, self)
        self.router = router
        self.router.register(self)

    def queue_execution(self, execution: CapabilityExecutionIF):
        """
        Queue this execution, or else put it in the queue (if it needs to run a PrepareAndRunWorkflow)

        :param execution:
        """
        # mark this execution as queued
        execution.queue_state = QueueState.Queued.name
        self.capability_info.save_execution(execution)

        # now tell the queue manager to handle it, once we're done with everything else
        transaction.get().addAfterCommitHook(lambda _: transaction.begin() and self.queue_manager.execution_requested())

    def resume_execution(self, execution: CapabilityExecutionIF):
        """
        Called by the queue manager when the execution can resume
        :param execution:  the execution whose time has come
        """
        # Mark this execution as no longer queued
        execution = self.capability_info.lookup_execution(execution.id)
        execution.queue_state = QueueState.Running.name
        self.capability_info.save_execution(execution)
        subject = execution.__json__()

        def send_message(status, execution):
            self.messenger.send_message(type="resume", service="capability", subject=execution)

        transaction.get().addAfterCommitHook(send_message, (subject,))
        transaction.commit()

    def execute(self, execution: CapabilityExecutionIF) -> CapabilityExecutionIF:
        """
        Run a capability with the requested settings

        :param execution: Capability execution
        :return: Execution of the request
        """
        request = execution.capability_request
        request.state = CapabilityRequestState.Submitted.name
        self.capability_info.save_entity(request)
        return self.start(execution)

    def start(self, execution: CapabilityExecutionIF):
        self.messenger.send_message(type="start", service="capability", subject=execution.__json__())
        return execution

    @on_message(service="workflow", type="workflow-complete")
    @on_message(service="workflow", type="workflow-failed")
    def mark_execution_as_not_running(self, **message):
        # As a friendly service, we update the queue state here regardless of whatever other processing occurs
        if execution := self.execution_from_message(**message):
            execution.queue_state = QueueState.NotQueued.name
            self.capability_info.save_execution(execution)
            self.queue_manager.execution_complete()

    @on_message(service="capability")
    @on_message(service="workflow")
    def process_message(self, **message: Dict):
        """
        Advance execution that is waiting to its next state

        :param message: Workflow-execution message
        """
        logger.info(f"In process_message, received message: {message}")

        # we ignore a few types of message that we mostly generate
        if message["type"] == "capability-complete" or message["type"] == "execution-complete":
            return

        try:
            if execution := self.execution_from_message(**message):
                logger.info("Found execution %s", execution.id)
                execution.signal(self, message)
                self.capability_info.save_execution(execution)
                self.check_for_complete(execution)
                transaction.commit()
            else:
                logger.info("Did not find an execution in this message somehow")
        except Exception as e:
            logger.exception("Something went wrong")

    def submit_workflow_request(
        self,
        execution_id: int,
        workflow_name: str,
        workflow_args: dict,
        files: List[AbstractFile],
    ):
        """
        Submit a workflow request to the workflow service

        :param execution_id:
        :param workflow_name:
        :param workflow_args:
        :param files:
        """
        logger.info("Submitting workflow request")
        transfer_files_string = self._get_transfer_files_property(files)

        args_list = {**workflow_args, "files_to_transfer": transfer_files_string}

        logger.info(
            "creating workflow request with name %s, args %s (and arg type %s)",
            workflow_name,
            args_list,
            type(args_list),
        )
        workflow_request = self.workflow_service.create_workflow_request(workflow=workflow_name, argument=args_list)

        logger.info(
            "My workflow_request is %s, with type %s and workflow request ID %s",
            workflow_request,
            type(workflow_request),
            workflow_request.workflow_request_id,
        )

        for file in files:
            self.workflow_service.attach_file_to_request(workflow_request, file.filename, file.content)
            logger.info(f"Added file {file.filename} to workflow request #{workflow_request.workflow_request_id}")

        execution = self.capability_info.lookup_execution(execution_id)
        execution.current_workflow_request_id = workflow_request.workflow_request_id
        self.capability_info.save_execution(execution)

        def execute_workflow_request(status, request):
            logger.info("Attempting to execute workflow request %s", request)
            self.workflow_service.execute(request)

        transaction.get().addAfterCommitHook(execute_workflow_request, (workflow_request,))

    def execution_from_message(self, **message: Dict) -> Optional[CapabilityExecutionIF]:
        """
        Extract a CapabilityExecutionIF from a message

        :param message: Message
        :return: CapabilityExecution if ID is found in message; else None
        """
        # first, let's check for a capability execution ID
        if message["subject"]["type"] == "CapabilityExecution":
            # if we found one, we're basically done
            return self.capability_info.lookup_execution(message["subject"]["id"])
        elif message["subject"]["type"] == "CapabilityRequest":
            return self.capability_info.lookup_execution(message["subject"]["current_execution"]["id"])
        elif message["service"] == "workflow":
            # OK, the message is from the workflow layer, so we need to look up using the workflow request ID instead
            logger.info(
                "Trying to find an execution with workflow_request_id=%s", message["subject"]["workflow_request_id"]
            )
            return self.capability_info.lookup_execution_by_workflow_request_id(
                message["subject"]["workflow_request_id"]
            )
        else:
            logger.error("Error: unable to ascertain the capability execution for message %s", message)
            return None

    def _complete_execution(self, execution: CapabilityExecutionIF) -> CapabilityExecutionIF:
        """
        Perform necessary steps to declare a capability execution complete

        :param execution: Execution to complete
        :return: Execution with updated metadata
        """
        execution.queue_state = QueueState.NotQueued.name

        execution_complete_msg = CapabilityMessageArchitect(execution=execution).compose_message("execution_complete")
        self.messenger.send_message(**execution_complete_msg)

        return execution

    def _fail_execution(self, execution: CapabilityExecutionIF) -> CapabilityExecutionIF:
        """
        Perform necessary steps to declare a capability execution failed

        :param execution: Execution to fail
        :return: Execution with updated metadata
        """
        execution.queue_state = QueueState.NotQueued.name

        execution_complete_msg = CapabilityMessageArchitect(execution=execution).compose_message("execution_failed")
        self.messenger.send_message(**execution_complete_msg)

        return execution

    def check_for_complete(self, execution: CapabilityExecutionIF):
        if execution.state.final():
            if execution.state == "Complete":
                self._complete_execution(execution)
            elif execution.state == "Failed":
                self._fail_execution(execution)

    @staticmethod
    def _get_transfer_files_property(files: List[AbstractFile]):
        separator = ", ./"
        return separator + separator.join(file.filename for file in files)
